@@ -44,6 +44,7 @@
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
+#include <dpkg/string.h>
 #include <dpkg/path.h>
 #include <dpkg/dir.h>
 #include <dpkg/fdio.h>
@@ -104,9 +105,11 @@ ensure_package_clientdata(struct pkginfo *pkg)
   pkg->clientdata = nfmalloc(sizeof(struct perpackagestate));
   pkg->clientdata->istobe = PKG_ISTOBE_NORMAL;
   pkg->clientdata->color = PKG_CYCLE_WHITE;
+  pkg->clientdata->enqueued = false;
   pkg->clientdata->fileslistvalid = false;
   pkg->clientdata->files = NULL;
   pkg->clientdata->replacingfilesandsaid = 0;
+  pkg->clientdata->cmdline_seen = 0;
   pkg->clientdata->listfile_phys_offs = 0;
   pkg->clientdata->trigprocdeferred = NULL;
 }
@@ -309,8 +312,10 @@ pkg_sorter_by_listfile_phys_offs(const void *a, const void *b)
    * INT_MAX. */
   if (pa->clientdata->listfile_phys_offs < pb->clientdata->listfile_phys_offs)
     return -1;
-  else
+  else if (pa->clientdata->listfile_phys_offs > pb->clientdata->listfile_phys_offs)
     return 1;
+  else
+    return 0;
 }
 
 static void
@@ -523,9 +528,9 @@ struct fileiterator {
   int nbinn;
 };
 
-/* This must always be a power of two. If you change it consider changing
- * the per-character hashing factor (currently 1785 = 137 * 13) too. */
-#define BINS (1 << 17)
+/* This must always be a prime for optimal performance.
+ * This is the closest one to 2^18 (262144). */
+#define BINS 262139
 
 static struct filenamenode *bins[BINS];
 
@@ -576,12 +581,6 @@ void filesdbinit(void) {
     }
 }
 
-static int hash(const char *name) {
-  int v= 0;
-  while (*name) { v *= 1787; v += *name; name++; }
-  return v;
-}
-
 struct filenamenode *findnamenode(const char *name, enum fnnflags flags) {
   struct filenamenode **pointerp, *newnode;
   const char *orig_name = name;
@@ -590,7 +589,7 @@ struct filenamenode *findnamenode(const char *name, enum fnnflags flags) {
    * leading slash. */
   name = path_skip_slash_dotslash(name);
 
-  pointerp= bins + (hash(name) & (BINS-1));
+  pointerp = bins + (str_fnv_hash(name) % (BINS));
   while (*pointerp) {
     /* XXX: Why is the assert needed? It's checking already added entries. */
     assert((*pointerp)->name[0] == '/');
